
#define EX_UTILS_NO_FUNCS 1
#include "ex_utils.h"

#include "mk.h"

#include "audispd.h"

Opt_serv_policy_opts *audispd_policy_make(Opt_serv_opts *beg)
{
  Audispd_policy_opts *opts = MK(sizeof(Audispd_policy_opts));
  Vstr_ref *ref = NULL;
  
  if (!opts)
    goto mk_opts_fail;
  
  if (!(ref = vstr_ref_make_ptr(opts, opt_policy_free)))
    goto mk_ref_fail;
  opts->s->ref = ref;

  if (!opt_policy_init(beg, opts->s))
    goto policy_init_fail;

  /* audispd_policy_init() */
  opts->max_msgs = AUDISPD_CONF_MAX_MSGS;
  opts->max_data = AUDISPD_CONF_MAX_DATA;
  
  opts->limit_reached = AUDISPD_CONF_TYPE_LIM_REACHED;
  
  opts->managed_con_died = AUDISPD_CONF_TYPE_MAN_CON_DIED;
  
  opts->filts    = NULL;
  opts->filt_sz  = 0;
  opts->filt_num = 0;
  
  return (opts->s);

 policy_init_fail:
  ref->ptr = NULL;
  vstr_ref_del(ref);
 mk_ref_fail:
  F(opts);
 mk_opts_fail:
  return (NULL);
}

void audispd_policy_change_con(struct Con *con, Audispd_policy_opts *policy)
{
  opt_policy_change_con(con->evnt, policy->s);
  con->policy = policy;
}

int audispd_policy_copy(Opt_serv_policy_opts *passed_dst,
                        const Opt_serv_policy_opts *passed_src)
{
  Audispd_policy_opts *dst = (Audispd_policy_opts *)passed_dst;
  const Audispd_policy_opts *src = (const Audispd_policy_opts *)passed_src;

  dst->max_msgs = src->max_msgs;
  dst->max_data = src->max_data;

  dst->limit_reached = src->limit_reached;
  dst->managed_con_died = src->managed_con_died;

  if (dst->filt_sz < src->filt_num)
  {
    F(dst->filts);
    dst->filt_sz  = 0;
    dst->filt_num = 0;
    if (!(dst->filts = MK(sizeof(int) * src->filt_num)))
      return (FALSE);
    dst->filt_sz  = src->filt_num;
  }
  dst->filt_num = src->filt_num;
  memcpy(dst->filts, src->filts, sizeof(int) * src->filt_num);
  
  return (TRUE);
}

int  audispd_conf_init(Audispd_opts *audispd_opts)
{
  unsigned int scan = 0;
  
  if (!opt_serv_conf_init(audispd_opts->s, AF_LOCAL))
    goto opts_init_fail;
  
  if (!(audispd_opts->main_input = vstr_make_base(NULL)))
    goto audispd_init_fail;

  while (scan < AUDISPD_CONF_MAX_CMDS)
  {
    Audispd_cmd *cmd = &audispd_opts->cmds[scan++];
    unsigned int ascan = 0;
    
    if (!(cmd->name = vstr_make_base(NULL)))
      goto audispd_init_fail;  
    if (!(cmd->path = vstr_make_base(NULL)))
      goto audispd_init_fail;

    while (ascan < AUDISPD_CONF_MAX_CMD_ARGS)
      if (!(cmd->args[ascan++] = vstr_make_base(NULL)))
        goto audispd_init_fail;
    
    if (!(cmd->def_policy = vstr_make_base(NULL)))
      goto audispd_init_fail;
  }
  
  return (TRUE);
  
 audispd_init_fail:
  audispd_conf_free(audispd_opts);
 opts_init_fail:
  return (FALSE);
}

void audispd_conf_free(Audispd_opts *audispd_opts)
{
  Opt_serv_policy_opts *scan = audispd_opts->s->def_policy;
  unsigned int cscan = 0;
  
  while (scan)
  {
    Audispd_policy_opts *tmp = (Audispd_policy_opts *)scan;
    Opt_serv_policy_opts *scan_next = scan->next;

    F(tmp->filts);
    
    scan = scan_next;
  }

  while (cscan < AUDISPD_CONF_MAX_CMDS)
  {
    Audispd_cmd *cmd = &audispd_opts->cmds[cscan++];
    unsigned int ascan = 0;
    
    vstr_free_base(cmd->name);
    vstr_free_base(cmd->path);

    while (ascan < AUDISPD_CONF_MAX_CMD_ARGS)
      vstr_free_base(cmd->args[ascan++]);
    
    vstr_free_base(cmd->def_policy);
  }

  opt_policy_sc_all_ref_del(audispd_opts->s);

  vstr_free_base(audispd_opts->main_input);
  
  conf_parse_free(audispd_opts->conf); audispd_opts->conf = NULL;

  opt_serv_conf_free_beg(audispd_opts->s);
  opt_serv_conf_free_end(audispd_opts->s);
}

extern int audit_name_to_msg_type(const char *); /* FIXME: hack */
static int audispd__conf_main_policy_d1(Audispd_policy_opts *opts,
                                        Conf_parse *conf, Conf_token *token,
                                        int clist)
{
  if (0) { }
  
  else if (OPT_SERV_SYM_EQ("match-init"))
    OPT_SERV_SC_MATCH_INIT(opts->s->beg,
                           audispd__conf_main_policy_d1(opts, conf, token,
                                                        clist));

  else if (OPT_SERV_SYM_EQ("limit"))
  {
    CONF_SC_MAKE_CLIST_BEG(limit, clist);
    
    else if (FALSE && (OPT_SERV_SYM_EQ("msgs") || OPT_SERV_SYM_EQ("messages")))
      OPT_SERV_X_UINT(opts->max_msgs); /* don't have the data atm. */
    else if (OPT_SERV_SYM_EQ("data"))
      OPT_SERV_X_UINT(opts->max_data);
    
    CONF_SC_MAKE_CLIST_END();
  }
    
  else if (OPT_SERV_SYM_EQ("limit-reached"))
  {
    CONF_SC_PARSE_TOP_TOKEN_RET(conf, token, FALSE);
    
    if (FALSE) { }
    
    else if (OPT_SERV_SYM_EQ("drop-messages") ||
             OPT_SERV_SYM_EQ("drop-msgs") || OPT_SERV_SYM_EQ("drop"))
      opts->limit_reached = AUDISPD_TYPE_LIM_REACHED_DROP;
    else if (OPT_SERV_SYM_EQ("close-connection") || OPT_SERV_SYM_EQ("close"))
      opts->limit_reached = AUDISPD_TYPE_LIM_REACHED_CLOSE;
    else if (FALSE &&
             (OPT_SERV_SYM_EQ("stop-input") || OPT_SERV_SYM_EQ("stop")))
      opts->limit_reached = AUDISPD_TYPE_LIM_REACHED_STOP;
    
    else
      return (FALSE);
  }
  
  else if (OPT_SERV_SYM_EQ("filter"))
  {
    CONF_SC_MAKE_CLIST_BEG(filt, clist);
    
    else if (OPT_SERV_SYM_EQ("audit-messages-type") ||
             OPT_SERV_SYM_EQ("audit-msg-type"))
    {
      CONF_SC_MAKE_CLIST_BEG(filt_aud_msg, clist);
      
      else if (!clist && OPT_SERV_SYM_EQ("<none>"))
        opts->filt_num = 0;
      else if (OPT_SERV_SYM_EQ("+=") || OPT_SERV_SYM_EQ("accept"))
      {
        const Vstr_sect_node *val  = NULL;
        const char *cstr = NULL;
        int id = 0;
        
        CONF_SC_MAKE_CLIST_BEG(filt_aud_msg_add, clist);
      
        else if (clist)
          return (FALSE);

        else if (!(val = conf_token_value(token)))
          return (FALSE);
        
        else if (!(cstr = vstr_export_cstr_ptr(conf->data, val->pos, val->len)))
          return (FALSE);
        
        else if ((id = audit_name_to_msg_type(cstr)))
        {
          if (!opts->filt_sz)
          {
            if (!(opts->filts = MK(sizeof(int) * 2)))
              return (FALSE);
            opts->filt_sz = 2;
          }
          else if (opts->filt_num >= opts->filt_sz)
          {
            int *tmp = NULL;
            
            if (!MV(opts->filts, tmp, sizeof(int) * (opts->filt_sz + 2)))
              return (FALSE);
            opts->filt_sz += 2;
          }
          
          opts->filts[opts->filt_num++] = id;
        }
        
        CONF_SC_MAKE_CLIST_END();
      }

      CONF_SC_MAKE_CLIST_END();
    }

    CONF_SC_MAKE_CLIST_END();
  }
  
  else if (OPT_SERV_SYM_EQ("managed-connection"))
  {
    CONF_SC_MAKE_CLIST_BEG(man_con, clist);
    
    else if (OPT_SERV_SYM_EQ("died"))
    {
      CONF_SC_PARSE_TOP_TOKEN_RET(conf, token, FALSE);
    
      if (FALSE) { }
      
      else if (OPT_SERV_SYM_EQ("restart"))
        opts->managed_con_died = AUDISPD_TYPE_MAN_CON_DIED_RESTART;
      else if (OPT_SERV_SYM_EQ("ignore"))
        opts->managed_con_died = AUDISPD_TYPE_MAN_CON_DIED_IGNORE;
      else if (OPT_SERV_SYM_EQ("cascade"))
        opts->managed_con_died = AUDISPD_TYPE_MAN_CON_DIED_CASCADE;
      
      else
        return (FALSE);
    }

    CONF_SC_MAKE_CLIST_END();
  }
  
  else
    return (FALSE);
  
  return (TRUE);
}

static int audispd__conf_main_policy(Audispd_opts *opts,
                                     Conf_parse *conf, Conf_token *token)
{
  Opt_serv_policy_opts *popts = NULL;
  Conf_token *ntoken = NULL;
  unsigned int cur_depth = opt_policy_sc_conf_parse(opts->s, conf, token,
                                                    &popts, &ntoken);
  
  if (!cur_depth)
    return (FALSE);

  do
  {
    int clist = FALSE;
    
    CONF_SC_MAKE_CLIST_MID(cur_depth, clist);
    
    else if (audispd__conf_main_policy_d1((Audispd_policy_opts *)popts,
                                          conf, token, clist))
    { }
    
    CONF_SC_MAKE_CLIST_END();
  } while (ntoken &&
           (cur_depth = opt_policy_sc_conf_parse(opts->s, conf, token,
                                                 &popts, &ntoken)));
  return (TRUE);
}

static int audispd__conf_main_d1(Audispd_opts *audispd_opts,
                                 Conf_parse *conf, Conf_token *token, int clist)
{
  if (OPT_SERV_SYM_EQ("org.and.daemon-conf-1.0"))
  {
    if (!opt_serv_conf(audispd_opts->s, conf, token))
      return (FALSE);
  }
  else if (OPT_SERV_SYM_EQ("match-init"))
    OPT_SERV_SC_MATCH_INIT(audispd_opts->s,
                           audispd__conf_main_d1(audispd_opts,
                                                 conf, token, clist));
  
  else if (OPT_SERV_SYM_EQ("main-input"))
  {
    return (opt_serv_sc_make_static_input(audispd_opts->s, conf, token,
                                          audispd_opts->main_input));
  }
  
  else if (OPT_SERV_SYM_EQ("managed-connection"))
  {
    Audispd_cmd *cmd = NULL;
    unsigned int scan = 0;

    OPT_SERV_X_SINGLE_VSTR(conf->tmp);

    while (scan < audispd_opts->cmd_num)
    {
      cmd = &audispd_opts->cmds[scan++];
      if (vstr_cmp_eq(conf->tmp, 1, conf->tmp->len,
                      cmd->name, 1, cmd->name->len))
        break;
      cmd = NULL;
    }
    if (!cmd && (audispd_opts->cmd_num < AUDISPD_CONF_MAX_CMDS))
    {
      cmd = &audispd_opts->cmds[audispd_opts->cmd_num++];

      ASSERT(!cmd->name->len);
      if (!vstr_add_vstr(cmd->name, 0, conf->tmp, 1, conf->tmp->len,
                         VSTR_TYPE_ADD_BUF_REF))
        return (FALSE);
    }
  
    if (!cmd)
      return (FALSE);
    
    CONF_SC_MAKE_CLIST_BEG(managed_prog_con, clist);
    
    else if (OPT_SERV_SYM_EQ("path"))
    {
      if (!opt_serv_sc_make_static_input(audispd_opts->s, conf, token,
                                         cmd->path))
        return (FALSE);
    }
    else if (OPT_SERV_SYM_EQ("args"))
    {
      CONF_SC_MAKE_CLIST_BEG(args, clist);

      else if (!clist)
      {
        cmd->arg_num = 1;
        OPT_SERV_X_VSTR(audispd_opts->s, cmd->args[0]);
      }

      else if (TRUE)
      {
        cmd->arg_num = 0;
        
        token->num--; /* FIXME: massive hack */
        
        while ((cmd->arg_num < AUDISPD_CONF_MAX_CMD_ARGS) &&
               conf_token_list_num(token, token->depth_num))
        {
          Vstr_base *tmp = cmd->args[cmd->arg_num++];
          OPT_SERV_X_VSTR(audispd_opts->s, tmp);
        }
      }
      
      CONF_SC_MAKE_CLIST_END();
    }
    else if (OPT_SERV_SYM_EQ("policy"))
      OPT_SERV_X_VSTR(audispd_opts->s, cmd->def_policy);
      
    CONF_SC_MAKE_CLIST_END();
  }
  
  else if (OPT_SERV_SYM_EQ("policy"))
  {
    if (!audispd__conf_main_policy(audispd_opts, conf, token))
      return (FALSE);
  }
  
  /*  else if (OPT_SERV_SYM_EQ("match-connection"))
  {
    if (!audispd__match_make(conf, token, AUDISPD_CONF_MAIN_MATCH_CON,
                             audispd_opts->match_connection,
                             audispd_opts->tmp_match_connection))
      return (FALSE);
    
    if (!conf_parse_end_token(conf, token, conf_token_at_depth(token)))
      return (FALSE);
  }
  */
  
  else
    return (FALSE);
  
  return (TRUE);  
}


static int audispd_conf_main(Audispd_opts *opts,
                             Conf_parse *conf, Conf_token *token)
{
  unsigned int cur_depth = token->depth_num;
  int clist = FALSE;
  
  if (!OPT_SERV_SYM_EQ("com.redhat.audispd-conf-main-1.0"))
    return (FALSE);

  CONF_SC_MAKE_CLIST_MID(cur_depth, clist);

  else if (audispd__conf_main_d1(opts, conf, token, clist))
  { }
  
  CONF_SC_MAKE_CLIST_END();
  
  /* And they all live together ... dum dum */
  if (conf->data->conf->malloc_bad)
    return (FALSE);
  
  return (TRUE);
}

#define AUDISPD_CONF__BEG_APP() do {                                    \
      ASSERT(!opts->conf_num || (conf->state == CONF_PARSE_STATE_END)); \
      prev_conf_num = conf->sects->num;                                 \
      /* reinit */                                                      \
      conf->state = CONF_PARSE_STATE_BEG;                               \
    } while (FALSE)

/* restore the previous parsing we've done and skip parsing it again */
#define AUDISPD_CONF__END_APP() do {                                    \
      if (opts->conf_num)                                               \
      {                                                                 \
        conf_parse_token(conf, token);                                  \
        conf_parse_num_token(conf, token, prev_conf_num);               \
      }                                                                 \
    } while (FALSE)

int audispd_conf_parse_cstr(Vstr_base *out, Audispd_opts *opts,const char *data)
{
  Conf_parse *conf    = opts->conf;
  Conf_token token[1] = {CONF_TOKEN_INIT};
  unsigned int prev_conf_num = 0;
  size_t pos = 1;
  size_t len = 0;

  ASSERT(opts && data);

  if (!conf && !(conf = conf_parse_make(NULL)))
    goto conf_malloc_fail;

  pos = conf->data->len + 1;
  if (!vstr_add_cstr_ptr(conf->data, conf->data->len,
                         "(com.redhat.audispd-conf-main-1.0 "))
    goto read_malloc_fail;
  if (!vstr_add_cstr_ptr(conf->data, conf->data->len, data))
    goto read_malloc_fail;
  if (!vstr_add_cstr_ptr(conf->data, conf->data->len,
                         ")"))
    goto read_malloc_fail;
  len = vstr_sc_posdiff(pos, conf->data->len);

  AUDISPD_CONF__BEG_APP();
  
  if (!conf_parse_lex(conf, pos, len))
    goto conf_fail;

  AUDISPD_CONF__END_APP();

  if (!conf_parse_token(conf, token))
    goto conf_fail;

  if ((token->type != CONF_TOKEN_TYPE_CLIST) || (token->depth_num != 1))
    goto conf_fail;
    
  if (!conf_parse_token(conf, token))
    goto conf_fail;

  ASSERT(OPT_SERV_SYM_EQ("com.redhat.audispd-conf-main-1.0"));
  if (!audispd_conf_main(opts, conf, token))
    goto conf_fail;

  if (token->num != conf->sects->num)
    goto conf_fail;

  opts->conf = conf;
  opts->conf_num++;
  
  return (TRUE);
  
 conf_fail:
  conf_parse_backtrace(out, data, conf, token);
 read_malloc_fail:
  conf_parse_free(conf);
 conf_malloc_fail:
  return (FALSE);
}

int audispd_conf_parse_file(Vstr_base *out, Audispd_opts *opts,
                            const char *fname)
{
  Conf_parse *conf    = opts->conf;
  Conf_token token[1] = {CONF_TOKEN_INIT};
  unsigned int prev_conf_num = 0;
  size_t pos = 1;
  size_t len = 0;
  
  ASSERT(opts && fname);

  if (!conf && !(conf = conf_parse_make(NULL)))
    goto conf_malloc_fail;

  pos = conf->data->len + 1;
  if (!vstr_sc_read_len_file(conf->data, conf->data->len, fname, 0, 0, NULL))
    goto read_malloc_fail;
  len = vstr_sc_posdiff(pos, conf->data->len);

  AUDISPD_CONF__BEG_APP();
  
  if (!conf_parse_lex(conf, pos, len))
    goto conf_fail;

  AUDISPD_CONF__END_APP();
  
  while (conf_parse_token(conf, token))
  {
    if ((token->type != CONF_TOKEN_TYPE_CLIST) || (token->depth_num != 1))
      goto conf_fail;
    
    if (!conf_parse_token(conf, token))
      goto conf_fail;
    
    if (OPT_SERV_SYM_EQ("org.and.daemon-conf-1.0"))
    {
      if (!opt_serv_conf(opts->s, conf, token))
        goto conf_fail;
    }
    else if (OPT_SERV_SYM_EQ("com.redhat.audispd-conf-main-1.0"))
    {
      if (!audispd_conf_main(opts, conf, token))
        goto conf_fail;
    }
    else
      goto conf_fail;
    
    opts->conf_num++;
  }

  opts->conf = conf;
  
  return (TRUE);
  
 conf_fail:
  conf_parse_backtrace(out, fname, conf, token);
  errno = 0;
 read_malloc_fail:
  if (errno && out) /* can't find config. file */
    vstr_add_fmt(out, out->len, "open(%s): %m", fname);
  
  conf_parse_free(conf);
 conf_malloc_fail:
  return (FALSE);
}

int audispd_sc_conf_parse_dir_file(Vstr_base *out, void *opts,
                                   const char *fname)
{
  return (audispd_conf_parse_file(out, opts, fname));
}
